/***************************************************************************
                          Qt.cpp  -  description
                             -------------------
    begin                : Fri Jul 4 2003
    copyright            : (C) 2003 by Richard Dale
    email                : Richard_Dale@tipitina.demon.co.uk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include <qglobal.h>
#include <qstring.h>
#include <qapplication.h>
#include <qmetaobject.h>
#include <private/qucomextra_p.h>

#undef DEBUG
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#ifndef __USE_POSIX
#define __USE_POSIX
#endif
#ifndef __USE_XOPEN
#define __USE_XOPEN
#endif
#ifdef _BOOL
#define HAS_BOOL
#endif

#include <ruby.h>

#ifndef QT_VERSION_STR
#define QT_VERSION_STR "Unknown"
#endif

#undef free
#undef malloc

#include "marshall.h"
#include "qtruby.h"
#include "smokeruby.h"

extern Smoke *qt_Smoke;
extern void init_qt_Smoke();
extern void smokeruby_free(void * ptr);

int do_debug = qtdb_none;

VALUE pointer_map = Qundef;
VALUE rv_qapp;
int object_count = 0;
void *_current_object = 0;    // TODO: ask myself if this is stupid

bool temporary_virtual_function_success = false;

QAsciiDict<Smoke::Index> methcache(2179);
QAsciiDict<Smoke::Index> classcache(2179);

VALUE rv_this = Qundef;
VALUE qt_module = Qundef;
VALUE qt_internal_module = Qundef;

Smoke::Index _current_object_class = 0;
Smoke::Index _current_method = 0;

/*
 * Type handling by moc is simple.
 *
 * If the type name matches /^(?:const\s+)?\Q$types\E&?$/, use the
 * static_QUType, where $types is join('|', qw(bool int double char* QString);
 *
 * Everything else is passed as a pointer! There are types which aren't
 * Smoke::tf_ptr but will have to be passed as a pointer. Make sure to keep
 * track of what's what.
 */

/*
 * Simply using typeids isn't enough for signals/slots. It will be possible
 * to declare signals and slots which use arguments which can't all be
 * found in a single smoke object. Instead, we need to store smoke => typeid
 * pairs. We also need additional informatation, such as whether we're passing
 * a pointer to the union element.
 */

enum MocArgumentType {
    xmoc_ptr,
    xmoc_bool,
    xmoc_int,
    xmoc_double,
    xmoc_charstar,
    xmoc_QString
};

struct MocArgument {
    // smoke object and associated typeid
    SmokeType st;
    MocArgumentType argType;
};


extern TypeHandler Qt_handlers[];
void install_handlers(TypeHandler *);

smokeruby_object *value_obj_info(VALUE ruby_value) {  // ptr on success, null on fail
	if (TYPE(ruby_value) != T_DATA) {
		return 0;
	}

	smokeruby_object * o = 0;	
	Data_Get_Struct(ruby_value, smokeruby_object, o);
    return o;
}

void *value_to_ptr(VALUE ruby_value) {  // ptr on success, null on fail
    smokeruby_object *o = value_obj_info(ruby_value);
    return o;
}

VALUE getPointerObject(void *ptr);

bool isQObject(Smoke *smoke, Smoke::Index classId) {
    if(!strcmp(smoke->classes[classId].className, "QObject"))
	return true;
    for(Smoke::Index *p = smoke->inheritanceList + smoke->classes[classId].parents;
	*p;
	p++) {
	if(isQObject(smoke, *p))
	    return true;
    }
    return false;
}

bool isDerivedFrom(Smoke *smoke, Smoke::Index classId, Smoke::Index baseId) {
    if(classId == baseId)
	return true;
    for(Smoke::Index *p = smoke->inheritanceList + smoke->classes[classId].parents;
	*p;
	p++) {
	if(isDerivedFrom(smoke, *p, baseId))
	    return true;
    }
    return false;
}

bool isDerivedFrom(Smoke *smoke, const char *className, const char *baseClassName) {
    if(!smoke || !className || !baseClassName)
	return false;
    Smoke::Index idClass = smoke->idClass(className);
    Smoke::Index idBase = smoke->idClass(baseClassName);
    return isDerivedFrom(smoke, idClass, idBase);
}

void unmapPointer(smokeruby_object *, Smoke::Index, void*);

VALUE getPointerObject(void *ptr) {
    VALUE hv = pointer_map;
    return rb_hash_aref(pointer_map, INT2NUM((int) ptr));
}

void unmapPointer(smokeruby_object *o, Smoke::Index classId, void *lastptr) {
    void *ptr = o->smoke->cast(o->ptr, o->classId, classId);
    VALUE hv = pointer_map;
    if(ptr != lastptr) {
	lastptr = ptr;
	VALUE key = INT2NUM((int) ptr);
//    HV *hv = pointer_map;
//    void *ptr = o->smoke->cast(o->ptr, o->classId, classId);
//    if(ptr != lastptr) {
//	lastptr = ptr;
//	VALUE keysv = newSViv((IV)ptr);
//	STRLEN len;
//	char *key = SvPV(keysv, len);
//	if(hv_exists(hv, key, len))
//	    hv_delete(hv, key, len, G_DISCARD);
//	SvREFCNT_dec(keysv);
    }
    for(Smoke::Index *i = o->smoke->inheritanceList + o->smoke->classes[classId].parents;
	*i;
	i++) {
	unmapPointer(o, *i, lastptr);
    }
}

// Store pointer in pointer_map hash : "pointer_to_Qt_object" => weak ref to associated Perl object
// Recurse to store it also as casted to its parent classes.

void mapPointer(VALUE obj, smokeruby_object *o, VALUE hv, Smoke::Index classId, void *lastptr) {
    void *ptr = o->smoke->cast(o->ptr, o->classId, classId);
    if(ptr != lastptr) {
	lastptr = ptr;
	VALUE key = INT2NUM((int) ptr);
//	STRLEN len;
//	char *key = SvPV(keysv, len);
//	VALUE rv = newSVsv(obj);
//	sv_rvweaken(rv);		// weak reference!
//	hv_store(hv, key, len, rv, 0);
	rb_hash_aset(hv, key, obj);
//	SvREFCNT_dec(keysv);
    }
    for(Smoke::Index *i = o->smoke->inheritanceList + o->smoke->classes[classId].parents;
	*i;
	i++) {
	mapPointer(obj, o, hv, *i, lastptr);
    }
}

Marshall::HandlerFn getMarshallFn(const SmokeType &type);

class VirtualMethodReturnValue : public Marshall {
    Smoke *_smoke;
    Smoke::Index _method;
    Smoke::Stack _stack;
    SmokeType _st;
    VALUE * _retval;
public:
    const Smoke::Method &method() { return _smoke->methods[_method]; }
    SmokeType type() { return _st; }
    Marshall::Action action() { return Marshall::FromVALUE; }
    Smoke::StackItem &item() { return _stack[0]; }
    VALUE * var() { return _retval; }
    void unsupported() {
	rb_bug("Cannot handle '%s' as return-type of virtual method %s::%s",
		type().name(),
		_smoke->className(method().classId),
		_smoke->methodNames[method().name]);
    }
    Smoke *smoke() { return _smoke; }
    void next() {}
    bool cleanup() { return false; }
    VirtualMethodReturnValue(Smoke *smoke, Smoke::Index meth, Smoke::Stack stack, VALUE * retval) :
	_smoke(smoke), _method(meth), _stack(stack), _retval(retval) {
	_st.set(_smoke, method().ret);
 	Marshall::HandlerFn fn = getMarshallFn(type());
	(*fn)(this);
   }
};

class VirtualMethodCall : public Marshall {
    Smoke *_smoke;
    Smoke::Index _method;
    Smoke::Stack _stack;
    VALUE _gv;
    int _cur;
    Smoke::Index *_args;
    VALUE *_sp;
    bool _called;
    VALUE _savethis;

public:
    SmokeType type() { return SmokeType(_smoke, _args[_cur]); }
    Marshall::Action action() { return Marshall::ToVALUE; }
    Smoke::StackItem &item() { return _stack[_cur + 1]; }
    VALUE * var() { return _sp + _cur; }
    const Smoke::Method &method() { return _smoke->methods[_method]; }
    void unsupported() {
	rb_bug("Cannot handle '%s' as argument of virtual method %s::%s",
		type().name(),
		_smoke->className(method().classId),
		_smoke->methodNames[method().name]);
    }
    Smoke *smoke() { return _smoke; }
    void callMethod() {
//	dSP;
	if(_called) return;
	_called = true;
//	SP = _sp + method().numArgs - 1;
//	PUTBACK;
//	int count = call_sv((SV*)GvCV(_gv), G_SCALAR);
//	SPAGAIN;
//	VirtualMethodReturnValue r(_smoke, _method, _stack, POPs);
//	PUTBACK;
//	FREETMPS;
//	LEAVE;
    }
    void next() {
	int oldcur = _cur;
	_cur++;
	while(!_called && _cur < method().numArgs) {
	    Marshall::HandlerFn fn = getMarshallFn(type());
//	    _sp[_cur] = sv_newmortal();
	    (*fn)(this);
	    _cur++;
	}
	callMethod();
	_cur = oldcur;
    }
    bool cleanup() { return false; }   // is this right?
    VirtualMethodCall(Smoke *smoke, Smoke::Index meth, Smoke::Stack stack, VALUE obj, VALUE gv) :
	_smoke(smoke), _method(meth), _stack(stack), _gv(gv), _cur(-1), _sp(0), _called(false) {
//	dSP;
//	ENTER;
//	SAVETMPS;
//	PUSHMARK(SP);
//	EXTEND(SP, method().numArgs);
//	_savethis = rv_this;
//	rv_this = newSVsv(obj);
//	_sp = SP + 1;
//	for(int i = 0; i < method().numArgs; i++)
//	    _sp[i] = sv_newmortal();
//	_args = _smoke->argumentList + method().args;
    }
    ~VirtualMethodCall() {
//	SvREFCNT_dec(rv_this);
//	rv_this = _savethis;
    }
};

class MethodReturnValue : public Marshall {
    Smoke *_smoke;
    Smoke::Index _method;
    VALUE * _retval;
    Smoke::Stack _stack;
public:
    MethodReturnValue(Smoke *smoke, Smoke::Index method, Smoke::Stack stack, VALUE * retval) :
	_smoke(smoke), _method(method), _retval(retval), _stack(stack) {
		Marshall::HandlerFn fn = getMarshallFn(type());
		(*fn)(this);
    }

    const Smoke::Method &method() { return _smoke->methods[_method]; }
    SmokeType type() { return SmokeType(_smoke, method().ret); }
    Marshall::Action action() { return Marshall::ToVALUE; }
    Smoke::StackItem &item() { return _stack[0]; }
    VALUE * var() {
    	return _retval;
    }
    void unsupported() {
	rb_bug("Cannot handle '%s' as return-type of %s::%s",
		type().name(),
		_smoke->className(method().classId),
		_smoke->methodNames[method().name]);
    }
    Smoke *smoke() { return _smoke; }
    void next() {}
    bool cleanup() { return false; }
};

class MethodCall : public Marshall {
    int _cur;
    Smoke *_smoke;
    Smoke::Stack _stack;
    Smoke::Index _method;
    Smoke::Index *_args;
    VALUE *_sp;
    int _items;
    VALUE _retval;
    bool _called;
public:
    MethodCall(Smoke *smoke, Smoke::Index method, VALUE *sp, int items) :
	_smoke(smoke), _method(method), _sp(sp), _items(items), _cur(-1), _called(false)
	{
		_args = _smoke->argumentList + _smoke->methods[_method].args;
		_items = _smoke->methods[_method].numArgs;
		_stack = new Smoke::StackItem[items + 1];
		_retval = Qundef;
    }

    ~MethodCall() {
		delete[] _stack;
//	SvREFCNT_dec(_retval);
    }

    SmokeType type() {
    	return SmokeType(_smoke, _args[_cur]);
    }

    Marshall::Action action() {
    	return Marshall::FromVALUE;
    }
    Smoke::StackItem &item() {
    	return _stack[_cur + 1];
    }

    VALUE * var() {
		if(_cur < 0) return &_retval;
//	SvGETMAGIC(*(_sp + _cur));
		return _sp + _cur;
    }
    
    inline const Smoke::Method &method() {
    	return _smoke->methods[_method];
    }
    
    void unsupported() {
		rb_bug("Cannot handle '%s' as argument to %s::%s",
			type().name(),
			_smoke->className(method().classId),
			_smoke->methodNames[method().name]);
    }
    
    Smoke *smoke() {
    	return _smoke;
    }
    
    inline void callMethod() {
		if(_called) return;
		_called = true;
		Smoke::ClassFn fn = _smoke->classes[method().classId].classFn;
		void *ptr = _smoke->cast(_current_object, _current_object_class, method().classId);
		_items = -1;
		(*fn)(method().method, ptr, _stack);
		MethodReturnValue r(_smoke, _method, _stack, &_retval);
    }
    
    void next() {
		int oldcur = _cur;
		_cur++;

		while(!_called && _cur < _items) {
	    	Marshall::HandlerFn fn = getMarshallFn(type());
	    	(*fn)(this);
	    	_cur++;
		}
		callMethod();
		_cur = oldcur;
    }
    
    bool cleanup() {
    	return true;
    }
};

class UnencapsulatedQObject : public QObject {
public:
    QConnectionList *public_receivers(int signal) const { return receivers(signal); }
    void public_activate_signal(QConnectionList *clist, QUObject *o) { activate_signal(clist, o); }
};

class EmitSignal : public Marshall {
    UnencapsulatedQObject *_qobj;
    int _id;
    MocArgument *_args;
    VALUE *_sp;
    int _items;
    int _cur;
    Smoke::Stack _stack;
    bool _called;
public:
    EmitSignal(QObject *qobj, int id, int items, MocArgument *args, VALUE *sp) :
	_qobj((UnencapsulatedQObject*)qobj), _id(id), _items(items), _args(args),
	_sp(sp), _cur(-1), _called(false) {
	_stack = new Smoke::StackItem[_items];
    }
    ~EmitSignal() {
	delete[] _stack;
    }
    const MocArgument &arg() { return _args[_cur]; }
    SmokeType type() { return arg().st; }
    Marshall::Action action() { return Marshall::FromVALUE; }
    Smoke::StackItem &item() { return _stack[_cur]; }
    VALUE * var() { return _sp + _cur; }
    void unsupported() {
	rb_bug("Cannot handle '%s' as signal argument", type().name());
    }
    Smoke *smoke() { return type().smoke(); }
    void emitSignal() {
	if(_called) return;
	_called = true;

	QConnectionList *clist = _qobj->public_receivers(_id);
	if(!clist) return;

	QUObject *o = new QUObject[_items + 1];
	for(int i = 0; i < _items; i++) {
	    QUObject *po = o + i + 1;
	    Smoke::StackItem *si = _stack + i;
	    switch(_args[i].argType) {
	      case xmoc_bool:
		static_QUType_bool.set(po, si->s_bool);
		break;
	      case xmoc_int:
		static_QUType_int.set(po, si->s_int);
		break;
	      case xmoc_double:
		static_QUType_double.set(po, si->s_double);
		break;
	      case xmoc_charstar:
		static_QUType_charstar.set(po, (char*)si->s_voidp);
		break;
	      case xmoc_QString:
		static_QUType_QString.set(po, *(QString*)si->s_voidp);
		break;
	      default:
		{
		    const SmokeType &t = _args[i].st;
		    void *p;
		    switch(t.elem()) {
		      case Smoke::t_bool:
			p = &si->s_bool;
			break;
		      case Smoke::t_char:
			p = &si->s_char;
			break;
		      case Smoke::t_uchar:
			p = &si->s_uchar;
			break;
		      case Smoke::t_short:
			p = &si->s_short;
			break;
		      case Smoke::t_ushort:
			p = &si->s_ushort;
			break;
		      case Smoke::t_int:
			p = &si->s_int;
			break;
		      case Smoke::t_uint:
			p = &si->s_uint;
			break;
		      case Smoke::t_long:
			p = &si->s_long;
			break;
		      case Smoke::t_ulong:
			p = &si->s_ulong;
			break;
		      case Smoke::t_float:
			p = &si->s_float;
			break;
		      case Smoke::t_double:
			p = &si->s_double;
			break;
		      case Smoke::t_enum:
			{
			    // allocate a new enum value
			    Smoke::EnumFn fn = SmokeClass(t).enumFn();
			    if(!fn) {
//				rb_warning("Unknown enumeration %s\n", t.name());
				p = new int((int)si->s_enum);
				break;
			    }
			    Smoke::Index id = t.typeId();
			    (*fn)(Smoke::EnumNew, id, p, si->s_enum);
			    (*fn)(Smoke::EnumFromLong, id, p, si->s_enum);
			    // FIXME: MEMORY LEAK
			}
			break;
		      case Smoke::t_class:
		      case Smoke::t_voidp:
			p = si->s_voidp;
			break;
		      default:
			p = 0;
			break;
		    }
		    static_QUType_ptr.set(po, p);
		}
	    }
	}

	_qobj->public_activate_signal(clist, o);
        delete[] o;
    }
    void next() {
	int oldcur = _cur;
	_cur++;

	while(!_called && _cur < _items) {
	    Marshall::HandlerFn fn = getMarshallFn(type());
	    (*fn)(this);
	    _cur++;
	}

	emitSignal();
	_cur = oldcur;
    }
    bool cleanup() { return true; }
};

class InvokeSlot : public Marshall {
    QObject *_qobj;
    VALUE _gv;
    int _items;
    MocArgument *_args;
    QUObject *_o;
    int _cur;
    bool _called;
    VALUE *_sp;
    Smoke::Stack _stack;
public:
    const MocArgument &arg() { return _args[_cur]; }
    SmokeType type() { return arg().st; }
    Marshall::Action action() { return Marshall::ToVALUE; }
    Smoke::StackItem &item() { return _stack[_cur]; }
    VALUE * var() { return _sp + _cur; }
    Smoke *smoke() { return type().smoke(); }
    bool cleanup() { return false; }
    void unsupported() {
	rb_bug("Cannot handle '%s' as slot argument\n", type().name());
    }
    void copyArguments() {
	for(int i = 0; i < _items; i++) {
	    QUObject *o = _o + i + 1;
	    switch(_args[i].argType) {
	      case xmoc_bool:
		_stack[i].s_bool = static_QUType_bool.get(o);
		break;
	      case xmoc_int:
		_stack[i].s_int = static_QUType_int.get(o);
		break;
	      case xmoc_double:
		_stack[i].s_double = static_QUType_double.get(o);
		break;
	      case xmoc_charstar:
		_stack[i].s_voidp = static_QUType_charstar.get(o);
		break;
	      case xmoc_QString:
		_stack[i].s_voidp = &static_QUType_QString.get(o);
		break;
	      default:	// case xmoc_ptr:
		{
		    const SmokeType &t = _args[i].st;
		    void *p = static_QUType_ptr.get(o);
		    switch(t.elem()) {
		      case Smoke::t_bool:
			_stack[i].s_bool = *(bool*)p;
			break;
		      case Smoke::t_char:
			_stack[i].s_char = *(char*)p;
			break;
		      case Smoke::t_uchar:
			_stack[i].s_uchar = *(unsigned char*)p;
			break;
		      case Smoke::t_short:
			_stack[i].s_short = *(short*)p;
			break;
		      case Smoke::t_ushort:
			_stack[i].s_ushort = *(unsigned short*)p;
			break;
		      case Smoke::t_int:
			_stack[i].s_int = *(int*)p;
			break;
		      case Smoke::t_uint:
			_stack[i].s_uint = *(unsigned int*)p;
			break;
		      case Smoke::t_long:
			_stack[i].s_long = *(long*)p;
			break;
		      case Smoke::t_ulong:
			_stack[i].s_ulong = *(unsigned long*)p;
			break;
		      case Smoke::t_float:
			_stack[i].s_float = *(float*)p;
			break;
		      case Smoke::t_double:
			_stack[i].s_double = *(double*)p;
			break;
		      case Smoke::t_enum:
			{
			    Smoke::EnumFn fn = SmokeClass(t).enumFn();
			    if(!fn) {
				rb_warning("Unknown enumeration %s\n", t.name());
				_stack[i].s_enum = *(int*)p;
				break;
			    }
			    Smoke::Index id = t.typeId();
			    (*fn)(Smoke::EnumToLong, id, p, _stack[i].s_enum);
			}
			break;
		      case Smoke::t_class:
		      case Smoke::t_voidp:
			_stack[i].s_voidp = p;
			break;
		    }
		}
	    }
	}
    }
    void invokeSlot() {
//	dSP;
	if(_called) return;
	_called = true;

//	SP = _sp + _items - 1;
//	PUTBACK;
//	int count = call_sv((SV*)GvCV(_gv), G_SCALAR);
//	SPAGAIN;
//	SP -= count;
//	PUTBACK;
//	FREETMPS;
//	LEAVE;
    }
    void next() {
	int oldcur = _cur;
	_cur++;

	while(!_called && _cur < _items) {
	    Marshall::HandlerFn fn = getMarshallFn(type());
	    (*fn)(this);
	    _cur++;
	}

	invokeSlot();
	_cur = oldcur;
    }
    InvokeSlot(QObject *qobj, VALUE gv, int items, MocArgument *args, QUObject *o) :
	_qobj(qobj), _gv(gv), _items(items), _args(args), _o(o), _cur(-1), _called(false) {
//	dSP;
//	ENTER;
//	SAVETMPS;
//	PUSHMARK(SP);
//	EXTEND(SP, items);
//	PUTBACK;
//	_sp = SP + 1;
//	for(int i = 0; i < _items; i++)
//	    _sp[i] = sv_newmortal();
	_stack = new Smoke::StackItem[_items];
	copyArguments();
    }
    ~InvokeSlot() {
	delete[] _stack;
    }

};

class QtRubySmokeBinding : public SmokeBinding {
public:
    QtRubySmokeBinding(Smoke *s) : SmokeBinding(s) {}
    void deleted(Smoke::Index classId, void *ptr) {
	VALUE obj = getPointerObject(ptr);
	smokeruby_object *o = value_obj_info(obj);
	if(do_debug & qtdb_gc) {
            printf("%p->~%s()\n", ptr, smoke->className(classId));
            if(o && o->ptr)
                object_count --;
	    //printf("Remaining objects: %d\n", object_count);
        }
	if(!o || !o->ptr) {
	    return;
	}
        unmapPointer(o, o->classId, 0);
	o->ptr = 0;
    }
    bool callMethod(Smoke::Index method, void *ptr, Smoke::Stack args, bool isAbstract) {
	VALUE obj = getPointerObject(ptr);
	smokeruby_object *o = value_obj_info(obj);
	if(do_debug & qtdb_virtual) fprintf(stderr, "virtual %p->%s::%s() called\n", ptr,
	    smoke->classes[smoke->methods[method].classId].className,
	    smoke->methodNames[smoke->methods[method].name]
        );

	if(!o) {
//	    if(!PL_dirty && (do_debug & qtdb_virtual) )   // if not in global destruction
//		fprintf(stderr, "Cannot find object for virtual method\n");
	    return false;
	}
//	HV *stash = SvSTASH(SvRV(obj));
//	if(*HvNAME(stash) == ' ')
//	    stash = gv_stashpv(HvNAME(stash) + 1, TRUE);
	const char *methodName = smoke->methodNames[smoke->methods[method].name];
	VALUE gv;
//	GV *gv = gv_fetchmethod_autoload(stash, methodName, 0);
//	if(!gv) return false;

	VirtualMethodCall c(smoke, method, args, obj, gv);
	// exception variable, just temporary
	temporary_virtual_function_success = true;
	c.next();
	bool ret = temporary_virtual_function_success;
	temporary_virtual_function_success = true;
	return ret;
    }
    char *className(Smoke::Index classId) {
	const char *className = smoke->className(classId);
	char *buf = new char[strlen(className) + 6];
	strcpy(buf, "Qt::");
	strcat(buf, className + 1);
	return buf;
    }
};

extern "C" {

// ----------------   Helpers -------------------

//---------- XS Autoload (for all functions except fully qualified statics & enums) ---------

VALUE catArguments(VALUE * sp, int n)
{
    VALUE r=rb_str_new2("");
//    for(int i = 0; i < n; i++) {
//        if(i) sv_catpv(r, ", ");
//        if(!SvOK(sp[i])) {
//            sv_catpv(r, "undef");
//        } else if(SvROK(sp[i])) {
//            smokeruby_object *o = value_obj_info(sp[i]);
//            if(o)
//                sv_catpv(r, o->smoke->className(o->classId));
//            else
//                sv_catsv(r, sp[i]);
//        } else {
//            bool isString = SvPOK(sp[i]);
//            STRLEN len;
//            char *s = SvPV(sp[i], len);
//            if(isString) sv_catpv(r, "'");
//            sv_catpvn(r, s, len > 10 ? 10 : len);
//            if(len > 10) sv_catpv(r, "...");
//            if(isString) sv_catpv(r, "'");
//        }
//    }
    return r;
}

Smoke::Index package_classid(const char *p)
{
     Smoke::Index *item = classcache.find(p);
     if(item)
         return *item;
     char *nisa = new char[strlen(p)+6];
     strcpy(nisa, p);
     strcat(nisa, "::ISA");
//     AV* isa=get_av(nisa, true);
     delete[] nisa;
//     for(int i=0; i<=av_len(isa); i++) {
//         SV** np = av_fetch(isa, i, 0);
//         if(np) {
//             Smoke::Index ix = package_classid(SvPV_nolen(*np));
//             if(ix) {
//                 classcache.insert(p, new Smoke::Index(ix));
//                 return ix;
//             }
//         }
//     }
     return (Smoke::Index) 0;
}

VALUE
set_obj_info(const char * className, smokeruby_object * o)
{
	VALUE klass = rb_funcall(	qt_internal_module,
								rb_intern("find_class"),
								1,
								rb_str_new2(className) );
	VALUE obj = Data_Wrap_Struct(klass, 0, smokeruby_free, (void *) o);
	return obj;
}

char *get_SVt(VALUE ruby_value)
{
    char *r = "";
    if(ruby_value == Qundef)
	r = "u";
    else if(TYPE(ruby_value) == T_FIXNUM || TYPE(ruby_value) == T_BIGNUM)
	r = "i";
    else if(TYPE(ruby_value) == T_FLOAT)
	r = "n";
    else if(TYPE(ruby_value) == T_STRING)
	r = "s";
    else if(TYPE(ruby_value) == T_DATA) {
	smokeruby_object *o = value_obj_info(ruby_value);
	if(!o) {
//            switch (svTYPE(SvRV(ruby_value))) {
//                case SVt_PVAV:
                  r = "a";
//                  break;
//                default:
//                  r = "r";
//            }
        }
	else
	    r = (char*)o->smoke->className(o->classId);
    }
    else
	r = "U";

    return r;
}

VALUE prettyPrintMethod(Smoke::Index id) {
    VALUE r = rb_str_new2("");
    Smoke::Method &meth = qt_Smoke->methods[id];
    const char *tname = qt_Smoke->types[meth.ret].name;
//    if(meth.flags & Smoke::mf_static) sv_catpv(r, "static ");
//    sv_catpvf(r, "%s ", (tname ? tname:"void"));
//    sv_catpvf(r, "%s::%s(", qt_Smoke->classes[meth.classId].className, qt_Smoke->methodNames[meth.name]);
//    for(int i = 0; i < meth.numArgs; i++) {
//        if(i) sv_catpv(r, ", ");
//        tname = qt_Smoke->types[qt_Smoke->argumentList[meth.args+i]].name;
//        sv_catpv(r, (tname ? tname:"void"));
//    }
//    sv_catpv(r, ")");
//    if(meth.flags & Smoke::mf_const) sv_catpv(r, " const");
    return r;
}

// --------------- Unary Keywords && Attributes ------------------


// implements unary 'this'
static VALUE
XS_this(VALUE self) {
//    ST(0) = rv_this;
    return INT2NUM(1);
}

// implements unary attributes: 'foo' means 'this->{foo}'
static VALUE
XS_attr(VALUE self) {
//    char *key = GvNAME(CvGV(cv));
//    U32 klen = strlen(key);
//    VALUE *svp = 0;
//    if(SvROK(rv_this) && SvTYPE(SvRV(rv_this)) == SVt_PVHV) {
//	HV *hv = (HV*)SvRV(rv_this);
//	svp = hv_fetch(hv, key, klen, 1);
//    }
//    if(svp) {
//	ST(0) = *svp;
//	return INT2NUM(1);
//    }
    return Qnil;
}

// implements unary SUPER attribute: 'SUPER' means ${(CopSTASH)::_INTERNAL_STATIC_}{SUPER}
static VALUE
XS_super(VALUE self) {
    char *key = "SUPER";
//    U32 klen = strlen(key);
    VALUE *svp = 0;
//    if(SvROK(rv_this) && SvTYPE(SvRV(rv_this)) == SVt_PVHV) {
//	HV *cs = (HV*)CopSTASH(PL_curcop);
//        if(!cs) return Qnil;
//        svp = hv_fetch(cs, "_INTERNAL_STATIC_", 17, 0);
//        if(!svp) return Qnil;
//        cs = GvHV((GV*)*svp);
//        if(!cs) return Qnil;
//        svp = hv_fetch(cs, "SUPER", 5, 0);
//    }
//    if(svp) {
//	ST(0) = *svp;
//	XSRETURN(1);
//    }
    return Qnil;
}

//---------- XS Autoload (for all functions except fully qualified statics & enums) ---------


bool avoid_fetchmethod = false;

static VALUE
method_missing(int argc, VALUE * argv, VALUE self)
{
	VALUE klass = rb_funcall(self, rb_intern("class"), 0);
	VALUE name = rb_funcall(klass, rb_intern("name"), 0);
    
	printf("In method_missing(argc: %d, argv[0]: %s TYPE: 0x%2.2x)\n",
			argc,
			rb_id2name( SYM2ID(argv[0]) ),
			TYPE(self) );

	VALUE * savestack = ALLOCA_N(VALUE, argc + 2);
	savestack[0] = rb_str_new2("Qt");
	savestack[1] = rb_str_new2(rb_id2name(SYM2ID(argv[0])));
	savestack[2] = name;
	for (int count = 1; count < argc; count++) {
		savestack[count+2] = argv[count];

	}

	rb_funcall2(qt_internal_module, rb_intern("do_method_missing"), argc+2, savestack);

        // FIXME: I shouldn't have to set the current object
        {
                smokeruby_object *o = value_obj_info(self);
                if(o && o->ptr) {
                    _current_object = o->ptr;
                    _current_object_class = o->classId;
                } else {
                    _current_object = 0;
                }
        }

	MethodCall c(qt_Smoke, _current_method, savestack+3, argc-1);
	c.next();
	VALUE result = *(c.var());
	return result;
}

static VALUE setThis(VALUE self, VALUE obj);

static VALUE
qt_new(int argc, VALUE * argv, VALUE klass)
{
	VALUE class_name = rb_funcall(klass, rb_intern("name"), 0);
	VALUE constructor_name = rb_str_new2("new");

	VALUE * savestack = ALLOCA_N(VALUE, argc + 3);
	savestack[0] = rb_str_new2("Qt");
	savestack[1] = constructor_name;
	savestack[2] = class_name;
	for (int count = 0; count < argc; count++) {
		savestack[count+3] = argv[count];

	}

	printf("In qt_new, argc: %d, self class_name: %s argv[0]: 0x%2.2x\n",
			argc,
			STR2CSTR(class_name),
			argv[0] );

	rb_funcall2(qt_internal_module, rb_intern("do_method_missing"), argc+3, savestack);
    
	// Success. Cache result.
	//methcache.insert((const char *)mcid, new Smoke::Index(_current_method));

        // FIXME: I shouldn't have to set the current object
        {
                smokeruby_object *o = value_obj_info(rv_this);
                if(o && o->ptr) {
                    _current_object = o->ptr;
                    _current_object_class = o->classId;
                } else {
                    _current_object = 0;
                }
        }

	MethodCall c(qt_Smoke, _current_method, savestack+3, argc);
	c.next();
	VALUE result = *(c.var());
	return result;
}

static VALUE
XS_AUTOLOAD(VALUE self) {
    // Err, XS autoload is borked. Lets try...
//    VALUE sv = get_sv("Qt::AutoLoad::AUTOLOAD", TRUE);
//    char *package = SvPV_nolen(sv);
//    char *method = 0;
//    for(char *s = package; *s; s++)
//	if(*s == ':') method = s;
//    if(!method) XSRETURN_NO;
//    *(method++ - 1) = 0;	// sorry for showing off. :)
//    int withObject = (*package == ' ') ? 1 : 0;
//    int isSuper = 0;
//    if(withObject) {
//        package++;
//        if(*package == ' ') {
//             isSuper = 1;
//             char *super = new char[strlen(package) + 7];
//             package++;
//             strcpy(super, package);
//             strcat(super, "::SUPER");
//             package = super;
//        }
//    }

//    HV *stash = gv_stashpv(package, TRUE);

//    if(do_debug & qtdb_autoload)
//        rb_warning("In XS Autoload for %s::%s()\n", package, method);

    // check for user-defined methods in the REAL stash; skip prefix
//    GV *gv = 0;
//    if(avoid_fetchmethod)
//	avoid_fetchmethod = false;
//    else
//	gv = gv_fetchmethod_autoload(stash, method, 0);

    // If we've made it here, we need to set rv_this
//    if(gv) {
//        if(do_debug & qtdb_autoload)
//            rb_warning("\tfound in %s's Perl stash\n", package);

	// call the defined Perl method with new 'this'
//	VALUE old_this;
//	if(withObject && !isSuper) {
//	    old_this = rv_this;
//	    rv_this = newSVsv(ST(0));
//	}

//	ENTER;
//	SAVETMPS;
//	PUSHMARK(SP - items + withObject);
//	PUTBACK;
//	int count = call_sv((SV*)GvCV(gv), G_SCALAR|G_EVAL);
//	SPAGAIN;
//	VALUE ret = newSVsv(TOPs);
//	SP -= count;
//	PUTBACK;
//	FREETMPS;
//	LEAVE;

//	if(withObject && !isSuper) {
//	    SvREFCNT_dec(rv_this);
//	    rv_this = old_this;
//	}
//        else if(isSuper)
//            delete[] package;

//        if(SvTRUE(ERRSV))
//            croak(SvPV_nolen(ERRSV));
//	ST(0) = sv_2mortal(ret);
//	XSRETURN(1);
//    }
//    else if(!strcmp(method, "DESTROY")) {
//        VALUE old_this;
//	if(withObject && !isSuper) {
//	    old_this = rv_this;
//	    rv_this = newSVsv(ST(0));
//	}
//	ENTER;
//	SAVETMPS;
//	PUSHMARK(SP);
//	PUSHs(sv_2mortal(newSVsv(rv_this)));
//	PUTBACK;
//	call_pv("Qt::_internal::destroy", G_DISCARD);
//	if(withObject && !isSuper) {
//	    SvREFCNT_dec(rv_this);
//	    rv_this = old_this;
//	}
//	FREETMPS;
//	LEAVE;

//    } else {

        // save the stack -- we'll need it
//        VALUE *savestack = new SV*[items+1];
//        VALUE saveobj = ST(0);
//        VALUE old_this;

//        Copy(SP - items + 1 + withObject, savestack, items-withObject, SV*);

        // Get the classid (eventually converting SUPER to the right Qt class)
//        Smoke::Index cid = package_classid(package);
        // Look in the cache
//        char *cname = (char*)qt_Smoke->className(cid);
//        QCString mcid(cname);
//        mcid += ';';
//        mcid += method;
//        for(int i=withObject ; i<items ; i++)
//        {
//            mcid += ';';
//            mcid += get_SVt(ST(i));
//        }
//        Smoke::Index *rcid = methcache.find((const char *)mcid);

//        if(rcid) {
            // Got a hit
//            _current_method = *rcid;
//            if(withObject && !isSuper) {
//                old_this = rv_this;
//                rv_this = newSVsv(ST(0));
//            }
//        }
//        else {

            // Find the C++ method to call. I'll do that from Perl for now

//            ENTER;
//            SAVETMPS;
//            PUSHMARK(SP - items + withObject);
//            EXTEND(SP, 3);
//            PUSHs(sv_2mortal(newSViv((IV)cid)));
//            PUSHs(sv_2mortal(newSVpv(method, 0)));
//            PUSHs(sv_2mortal(newSVpv(package, 0)));
//            PUTBACK;
//            if(withObject && !isSuper) {
//                old_this = rv_this;
//                rv_this = newSVsv(saveobj);
//            }
//            call_pv("Qt::_internal::do_autoload", G_DISCARD|G_EVAL);
//            FREETMPS;
//            LEAVE;

            // Restore rv_this on error, so that eval{ } works
//            if(SvTRUE(ERRSV)) {
//                if(withObject && !isSuper) {
//                        SvREFCNT_dec(rv_this);
//                        rv_this = old_this;
//                }
//                else if(isSuper)
//                        delete[] package;
//                delete[] savestack;
//                croak(SvPV_nolen(ERRSV));
//            }

            // Success. Cache result.
//            methcache.insert((const char *)mcid, new Smoke::Index(_current_method));
//        }

        // FIXME: I shouldn't have to set the current object
//        {
//                smokeruby_object *o = value_obj_info(rv_this);
//                if(o && o->ptr) {
//                    _current_object = o->ptr;
//                    _current_object_class = o->classId;
//                } else {
//                    _current_object = 0;
//                }
//        }
        // honor debugging channels
//        if(do_debug & qtdb_calls) {
//            rb_warning("Calling method\t%s\n", SvPV_nolen(sv_2mortal(prettyPrintMethod(_current_method))));
//            if(do_debug & qtdb_verbose)
//                rb_warning("with arguments (%s)\n", SvPV_nolen(sv_2mortal(catArguments(savestack, items-withObject))));
//        }
//        MethodCall c(qt_Smoke, _current_method, savestack, items-withObject);
//        c.next();
//        if(savestack)
//            delete[] savestack;

//        if(withObject && !isSuper) {
//                SvREFCNT_dec(rv_this);
//                rv_this = old_this;
//        }
//        else if(isSuper)
//                delete[] package;

//        VALUE ret = c.var();
//        SvREFCNT_inc(ret);
//        ST(0) = sv_2mortal(ret);
//        return INT2NUM(1);
//    }
//    if(isSuper)
//        delete[] package;
    return Qtrue;
}


//----------------- Sig/Slot ------------------


MocArgument *getmetainfo(VALUE gv, const char *name, int &offset, int &index, int &argcnt) {
//    char *signalname = GvNAME(gv);
//    HV *stash = GvSTASH(gv);

    // $meta = $stash->{META}
//    VALUE *svp = hv_fetch(stash, "META", 4, 0);
//    if(!svp) return 0;
//    HV *hv = GvHV((GV*)*svp);
//    if(!hv) return 0;

    // $metaobject = $meta->{object}
    // aka. Class->staticMetaObject
//    svp = hv_fetch(hv, "object", 6, 0);
//    if(!svp) return 0;
//    smokeruby_object *ometa = value_obj_info(*svp);
//    if(!ometa) return 0;
//    QMetaObject *metaobject = (QMetaObject*)ometa->ptr;

//    offset = metaobject->signalOffset();

    // $signals = $meta->{signal}
//    U32 len = strlen(name);
//    svp = hv_fetch(hv, name, len, 0);
//    if(!svp) return 0;
//    HV *signalshv = (HV*)SvRV(*svp);

    // $signal = $signals->{$signalname}
//    len = strlen(signalname);
//    svp = hv_fetch(signalshv, signalname, len, 0);
//    if(!svp) return 0;
//    HV *signalhv = (HV*)SvRV(*svp);

    // $index = $signal->{index}
//    svp = hv_fetch(signalhv, "index", 5, 0);
//    if(!svp) return 0;;
//    index = SvIV(*svp);

    // $argcnt = $signal->{argcnt}
//    svp = hv_fetch(signalhv, "argcnt", 6, 0);
//    if(!svp) return 0;
//    argcnt = SvIV(*svp);

    // $mocargs = $signal->{mocargs}
//    svp = hv_fetch(signalhv, "mocargs", 7, 0);
//    if(!svp) return 0;
//    MocArgument *args = (MocArgument*)SvIV(*svp);
    MocArgument *args = 0;

    return args;
}

MocArgument *getslotinfo(VALUE gv, int id, char *&slotname, int &index, int &argcnt, bool isSignal = false) {
//    HV *stash = GvSTASH(gv);

    // $meta = $stash->{META}
//    VALUE *svp = hv_fetch(stash, "META", 4, 0);
//    if(!svp) return 0;
//    HV *hv = GvHV((GV*)*svp);
//    if(!hv) return 0;

    // $metaobject = $meta->{object}
    // aka. Class->staticMetaObject
//    svp = hv_fetch(hv, "object", 6, 0);
 //   if(!svp) return 0;
//    smokeruby_object *ometa = value_obj_info(*svp);
//    if(!ometa) return 0;
//    QMetaObject *metaobject = (QMetaObject*)ometa->ptr;

//    int offset = isSignal ? metaobject->signalOffset() : metaobject->slotOffset();

//    index = id - offset;   // where we at
    // FIXME: make slot inheritance work
//    if(index < 0) return 0;
    // $signals = $meta->{signal}
//    const char *key = isSignal ? "signals" : "slots";
//    svp = hv_fetch(hv, key, strlen(key), 0);
//    if(!svp) return 0;
//    AV *signalsav = (AV*)SvRV(*svp);
//    svp = av_fetch(signalsav, index, 0);
//    if(!svp) return 0;
//    HV *signalhv = (HV*)SvRV(*svp);
    // $argcnt = $signal->{argcnt}
//    svp = hv_fetch(signalhv, "argcnt", 6, 0);
//    if(!svp) return 0;
//    argcnt = SvIV(*svp);
    // $mocargs = $signal->{mocargs}
//    svp = hv_fetch(signalhv, "mocargs", 7, 0);
//    if(!svp) return 0;
//    MocArgument *args = (MocArgument*)SvIV(*svp);
    MocArgument *args = 0;

//    svp = hv_fetch(signalhv, "name", 4, 0);
//    if(!svp) return 0;
//    slotname = SvPV_nolen(*svp);

    return args;
}

static VALUE
XS_signal(VALUE self) {
    smokeruby_object *o = value_obj_info(rv_this);
    QObject *qobj = (QObject*)o->smoke->cast(
	o->ptr,
	o->classId,
	o->smoke->idClass("QObject")
    );
    if(qobj->signalsBlocked()) return Qnil;

    int offset;
    int index;
    int argcnt;
    MocArgument *args = 0;

//    args = getmetainfo(CvGV(cv), "signal", offset, index, argcnt);
    if(!args) return Qnil;

    // Okay, we have the signal info. *whew*
//    if(items < argcnt)
//	croak("Insufficient arguments to emit signal");

//    EmitSignal signal(qobj, offset + index, argcnt, args, &ST(0));
//    signal.next();

    return Qnil;
}

static VALUE
XS_qt_invoke(VALUE self)
{
//    dXSARGS;
    // Arguments: int id, QUObject *o
//    int id = SvIV(ST(0));
//    QUObject *_o = (QUObject*)SvIV(SvRV(ST(1)));

    smokeruby_object *o = value_obj_info(rv_this);
    QObject *qobj = (QObject*)o->smoke->cast(
	o->ptr,
	o->classId,
	o->smoke->idClass("QObject")
    );

    // Now, I need to find out if this means me
    int index;
    char *slotname;
    int argcnt;
    MocArgument *args;
//    bool isSignal = !strcmp(GvNAME(CvGV(cv)), "qt_emit");
//    args = getslotinfo(CvGV(cv), id, slotname, index, argcnt, isSignal);
//    if(!args) {
	// throw an exception - evil style
//	temporary_virtual_function_success = false;
//	return Qnil;
//    }
//    HV *stash = GvSTASH(CvGV(cv));
//    GV *gv = gv_fetchmethod_autoload(stash, slotname, 0);
//    if(!gv) return Qnil;
//    InvokeSlot slot(qobj, gv, argcnt, args, _o);
//    slot.next();

    return self;
}

// --------------- Ruby C functions for Qt::_internal::* helpers  ----------------


static VALUE
getMethStat(VALUE self)
{
    VALUE result_list = rb_ary_new();
    rb_ary_push(result_list, INT2NUM((int)methcache.size()));
    rb_ary_push(result_list, INT2NUM((int)methcache.count()));
    return result_list;
}

static VALUE
getClassStat(VALUE self)
{
    VALUE result_list = rb_ary_new();
    rb_ary_push(result_list, INT2NUM((int)classcache.size()));
    rb_ary_push(result_list, INT2NUM((int)classcache.count()));
    return result_list;
}

static VALUE
getIsa(VALUE self, VALUE classId)
{
	VALUE parents_list = rb_ary_new();

    Smoke::Index *parents =
	qt_Smoke->inheritanceList +
	qt_Smoke->classes[NUM2INT(classId)].parents;

    while(*parents) {
//	printf("\tparent: %s\n", qt_Smoke->classes[*parents].className);
	rb_ary_push(parents_list, rb_str_new2(qt_Smoke->classes[*parents++].className));
//	XPUSHs(sv_2mortal(newSVpv(qt_Smoke->classes[*parents++].className, 0)));
	}
	return parents_list;
}


static VALUE
dontRecurse(VALUE self)
{
    avoid_fetchmethod = true;
    return self;
}

static VALUE
allocateMocArguments(VALUE self, VALUE count_value)
{
    int count = NUM2INT(count_value);
    return INT2NUM((long)new MocArgument[count + 1]);
}

static VALUE
setMocType(VALUE self, VALUE ptr_value, VALUE idx_value, VALUE name_value, VALUE static_type_value)
{
    void *ptr = (void *) NUM2LONG(ptr_value);
    int idx = NUM2INT(idx_value);
    char *name = STR2CSTR(name_value);
    char *static_type = STR2CSTR(static_type_value);
    Smoke::Index typeId = qt_Smoke->idType(name);
    if(!typeId) return Qfalse;
    MocArgument *arg = (MocArgument*)ptr;
    arg[idx].st.set(qt_Smoke, typeId);
    if(!strcmp(static_type, "ptr"))
	arg[idx].argType = xmoc_ptr;
    else if(!strcmp(static_type, "bool"))
	arg[idx].argType = xmoc_bool;
    else if(!strcmp(static_type, "int"))
	arg[idx].argType = xmoc_int;
    else if(!strcmp(static_type, "double"))
	arg[idx].argType = xmoc_double;
    else if(!strcmp(static_type, "char*"))
	arg[idx].argType = xmoc_charstar;
    else if(!strcmp(static_type, "QString"))
	arg[idx].argType = xmoc_QString;
    return Qtrue;
}

static VALUE
installsignal(VALUE self, VALUE name_value)
{
    char *name = STR2CSTR(name_value);
    char *file = __FILE__;
//    newXS(name, XS_signal, file);
	return self;
}

static VALUE
installqt_invoke(VALUE self, VALUE name_value)
{
    char *name = STR2CSTR(name_value);
    char *file = __FILE__;
//    newXS(name, XS_qt_invoke, file);
	return self;
}

static VALUE
setDebug(VALUE self, VALUE on_value)
{
    int on = NUM2INT(on_value);
    do_debug = on;
    return self;
}

#ifdef DEBUG
static VALUE
debug(VALUE self)
{
    return INT2NUM(do_debug);
}
#endif

static VALUE
getTypeNameOfArg(VALUE self, VALUE method_value, VALUE idx_value)
{
    int method = NUM2INT(method_value);
    int idx = NUM2INT(idx_value);
    Smoke::Method &m = qt_Smoke->methods[method];
    Smoke::Index *args = qt_Smoke->argumentList + m.args;
    return rb_str_new2((char*)qt_Smoke->types[args[idx]].name);
}

static VALUE
classIsa(VALUE self, VALUE className_value, VALUE base_value)
{
    char *className = STR2CSTR(className_value);
    char *base = STR2CSTR(base_value);
    return INT2NUM(isDerivedFrom(qt_Smoke, className, base));
}

static VALUE
insert_pclassid(VALUE self, VALUE p_value, VALUE ix_value)
{
    char *p = STR2CSTR(p_value);
    int ix = NUM2INT(ix_value);
    classcache.insert(p, new Smoke::Index((Smoke::Index)ix));
    return self;
}

static VALUE
find_pclassid(VALUE self, VALUE p_value)
{
    char *p = STR2CSTR(p_value);
    Smoke::Index *r = classcache.find(p);
    if(r)
        return INT2NUM((int)*r);
    else
        return INT2NUM(0);
}

static VALUE
insert_mcid(VALUE self, VALUE mcid_value, VALUE ix_value)
{
    char *mcid = STR2CSTR(mcid_value);
    int ix = NUM2INT(ix_value);
    methcache.insert(mcid, new Smoke::Index((Smoke::Index)ix));
    return self;
}

static VALUE
find_mcid(VALUE self, VALUE mcid_value)
{
    char *mcid = STR2CSTR(mcid_value);
    Smoke::Index *r = methcache.find(mcid);
    if(r)
        return INT2NUM((int)*r);
    else
        return INT2NUM(0);
}

static VALUE
getSVt(VALUE self, VALUE ruby_value)
{
//    get_SVt(sv);
    return self;
}

static VALUE
make_QUParameter(VALUE self, VALUE name_value, VALUE type_value, VALUE extra, VALUE inout_value)
{
    char *name = STR2CSTR(name_value);
    char *type = STR2CSTR(type_value);
    int inout = NUM2INT(inout_value);
    QUParameter *p = new QUParameter;
    p->name = new char[strlen(name) + 1];
    strcpy((char*)p->name, name);
    if(!strcmp(type, "bool"))
	p->type = &static_QUType_bool;
    else if(!strcmp(type, "int"))
	p->type = &static_QUType_int;
    else if(!strcmp(type, "double"))
	p->type = &static_QUType_double;
    else if(!strcmp(type, "char*") || !strcmp(type, "const char*"))
	p->type = &static_QUType_charstar;
    else if(!strcmp(type, "QString") || !strcmp(type, "QString&") ||
	    !strcmp(type, "const QString") || !strcmp(type, "const QString&"))
	p->type = &static_QUType_QString;
    else
	p->type = &static_QUType_ptr;
    // Lacking support for several types. Evil.
    p->inOut = inout;
    p->typeExtra = 0;
    return INT2NUM((long)p);
}

static VALUE
make_QMetaData(VALUE self, VALUE name_value, VALUE method_value)
{
    char *name = STR2CSTR(name_value);
    void *method = STR2CSTR(method_value);
    QMetaData *m = new QMetaData;		// will be deleted
    m->name = new char[strlen(name) + 1];
    strcpy((char*)m->name, name);
    m->method = (QUMethod*)method;
    m->access = QMetaData::Public;
    return INT2NUM((long) m);
}

static VALUE
make_QUMethod(VALUE self, VALUE name_value, VALUE params)
{
    char *name = STR2CSTR(name_value);
    QUMethod *m = new QUMethod;			// permanent memory allocation
    m->name = new char[strlen(name) + 1];	// this too
    strcpy((char*)m->name, name);
    m->count = 0;
    m->parameters = 0;
    /*
    if(SvOK(params) && SvRV(params)) {
	AV *av = (AV*)SvRV(params);
	m->count = av_len(av) + 1;
	if(m->count > 0) {
	    m->parameters = new QUParameter[m->count];
	    for(int i = 0; i < m->count; i++) {
		VALUE sv = av_shift(av);
		if(!SvOK(sv)) {
		    croak("Invalid paramater for QUMethod\n");
		QUParameter *p = (QUParameter*)SvIV(sv);
		SvREFCNT_dec(sv);
		((QUParameter*)m->parameters)[i] = *p;
		delete p;
	    }
	} else
	    m->count = 0;
    }
    */
    return INT2NUM((long) m);
}

static VALUE
make_QMetaData_tbl(VALUE self, VALUE list)
{
/*
    if(SvOK(list) && SvRV(list)) {
	AV *av = (AV*)SvRV(list);
	int count = av_len(av) + 1;
	QMetaData *m = new QMetaData[count];
	for(int i = 0; i < count; i++) {
	    VALUE sv = av_shift(av);
	    if(!SvOK(sv))
		croak("Invalid metadata\n");
	    QMetaData *old = (QMetaData*)SvIV(sv);
	    SvREFCNT_dec(sv);
	    m[i] = *old;
	    delete old;
	}
	return self; = (void*)m;
    }
 */
    return self;
}

static VALUE
make_metaObject(VALUE self, VALUE className_value, VALUE parent, VALUE slot_tbl_value, VALUE slot_count_value, VALUE signal_tbl_value, VALUE signal_count_value)
{
    char *className = STR2CSTR(className_value);
    void *slot_tbl = (void *) NUM2INT(slot_tbl_value);
    int slot_count = NUM2INT(slot_count_value);
    void *signal_tbl = (void *) NUM2INT(signal_tbl_value);
    int signal_count = NUM2INT(signal_count_value);
    smokeruby_object *po = value_obj_info(parent);
    if(!po || !po->ptr) {
    	rb_bug("Cannot create metaObject\n");
    }
    QMetaObject *meta = QMetaObject::new_metaobject(
	className, (QMetaObject*)po->ptr,
	(const QMetaData*)slot_tbl, slot_count,	// slots
	(const QMetaData*)signal_tbl, signal_count,	// signals
	0, 0,	// properties
	0, 0,	// enums
	0, 0);

    // this object-creation code is so, so wrong here
    VALUE hv = rb_ary_new();
//    VALUE obj = newRV_noinc((SV*)hv);

    smokeruby_object o;
    o.smoke = qt_Smoke;
    o.classId = qt_Smoke->idClass("QMetaObject");
    o.ptr = meta;
    o.allocated = true;
//    sv_magic((SV*)hv, 0, '~', (char*)&o, sizeof(o));
//    MAGIC *mg = mg_find((SV*)hv, '~');
//    mg->mg_virtual = &vtbl_smoke;
    char *buf = qt_Smoke->binding->className(o.classId);
//    sv_bless(obj, gv_stashpv(buf, TRUE));
    delete[] buf;
//    return obj;
    return self;
}

static VALUE
dumpObjects(VALUE self)
{
/*
    hv_iterinit(pointer_map);
    HE *e;
    while(e = hv_iternext(pointer_map)) {
	STRLEN len;
	VALUE sv = HeVAL(e);
	printf("key = %s, refcnt = %d, weak = %d, ref? %d\n", HePV(e, len), SvREFCNT(sv), SvWEAKREF(sv), SvROK(sv)?1:0);
	if(SvRV(sv))
	    printf("REFCNT = %d\n", SvREFCNT(SvRV(sv)));
	//SvREFCNT_dec(HeVAL(e));
	//HeVAL(e) = &PL_sv_undef;
    }
 */
	return self;
}

static VALUE
dangle(VALUE self, VALUE obj)
{
//    VALUE obj
//    if(SvRV(obj))
//	SvREFCNT_inc(SvRV(obj));
	return self;
}

static VALUE
setAllocated(VALUE self, VALUE obj, VALUE b_value)
{
    bool b = (bool) NUM2INT(b_value);
    smokeruby_object *o = value_obj_info(obj);
    if(o) {
	o->allocated = b;
    }
    return self;
}

static VALUE
setqapp(VALUE self, VALUE obj)
{
//    if(!obj || !SvROK(obj))
//        rb_bug("Invalid Qt::Application object. Couldn't set Qt::app()\n");
//    value_qapp = SvRV(obj);
    return self;
}

static VALUE
setThis(VALUE self, VALUE obj)
{
	rv_this = obj;
	return self;
}

static VALUE
deleteObject(VALUE self, VALUE obj)
{
    smokeruby_object *o = value_obj_info(obj);
    if(!o) { return Qnil; }
    QObject *qobj = (QObject*)o->smoke->cast(o->ptr, o->classId, o->smoke->idClass("QObject"));
    delete qobj;
    return self;
}

static VALUE
mapObject(VALUE self, VALUE obj)
{
    smokeruby_object *o = value_obj_info(obj);
    if(!o)
        return Qnil;
    SmokeClass c( o->smoke, o->classId );
    object_count ++;
    if(!c.hasVirtual() ) {
	return Qnil;
    }
    mapPointer(obj, o, pointer_map, o->classId, 0);
    return self;
}

static VALUE
isaQObject(VALUE self, VALUE obj)
{
    smokeruby_object *o = value_obj_info(obj);
    if(o && isQObject(o->smoke, o->classId)) {
		return INT2NUM(1);
    } else {
		return INT2NUM(0);
	}
}

static VALUE
isValidAllocatedPointer(VALUE self, VALUE obj)
{
    smokeruby_object *o = value_obj_info(obj);
    if(o && o->ptr && o->allocated) {
		return INT2NUM(1);
    } else {
		return INT2NUM(0);
	}
}

static VALUE
findAllocatedObjectFor(VALUE self, VALUE obj)
{
    smokeruby_object *o = value_obj_info(obj);
    VALUE ret;
    if(o && o->ptr && (ret = getPointerObject(o->ptr)))
        return ret;
    return Qnil;
}

static VALUE
idClass(VALUE self, VALUE name_value)
{
    char *name = STR2CSTR(name_value);
    return INT2NUM(qt_Smoke->idClass(name));
}

static VALUE
idMethodName(VALUE self, VALUE name_value)
{
    char *name = STR2CSTR(name_value);
    return INT2NUM(qt_Smoke->idMethodName(name));
}

static VALUE
idMethod(VALUE self, VALUE idclass_value, VALUE idmethodname_value)
{
    int idclass = NUM2INT(idclass_value);
    int idmethodname = NUM2INT(idmethodname_value);
    return INT2NUM(qt_Smoke->idMethod(idclass, idmethodname));
}

static VALUE
findMethod(VALUE self, VALUE c_value, VALUE name_value)
{
    char *c = STR2CSTR(c_value);
    char *name = STR2CSTR(name_value);
    VALUE result = Qundef;
    Smoke::Index meth = qt_Smoke->findMethod(c, name);
    printf("DAMNIT on %s::%s => %d\n", c, name, meth);
    if(!meth) {
	// empty list
    } else if(meth > 0) {
	Smoke::Index i = qt_Smoke->methodMaps[meth].method;
	if(!i) {		// shouldn't happen
	    rb_bug("Corrupt method %s::%s", c, name);
	} else if(i > 0) {	// single match
		result = INT2NUM(qt_Smoke->methodMaps[meth].method);
//	    PUSHs(sv_2mortal(newSViv(
//		(IV)qt_Smoke->methodMaps[meth].method
//	    )));
	} else {		// multiple match
	    i = -i;		// turn into ambiguousMethodList index
	    while(qt_Smoke->ambiguousMethodList[i]) {
		result = INT2NUM(qt_Smoke->methodMaps[meth].method);
    	printf("ambiguousMethodList %s::%s => %d\n", c, name, meth);
//		PUSHs(sv_2mortal(newSViv(
//		    (IV)qt_Smoke->ambiguousMethodList[i]
//		)));
		i++;
	    }
	}
    }
    return result;
}

static VALUE
findMethodFromIds(VALUE self, VALUE idclass_value, VALUE idmethodname_value)
{
    int idclass = NUM2INT(idclass_value);
    int idmethodname = NUM2INT(idmethodname_value);
    VALUE result = Qundef;
    Smoke::Index meth = qt_Smoke->findMethod(idclass, idmethodname);
    if(!meth) {
	// empty list
    } else if(meth > 0) {
	Smoke::Index i = qt_Smoke->methodMaps[meth].method;
	if(i >= 0) {	// single match
		result = INT2NUM((int) i);
//	    PUSHs(sv_2mortal(newSViv((IV)i)));
	} else {		// multiple match
	    i = -i;		// turn into ambiguousMethodList index
	    while(qt_Smoke->ambiguousMethodList[i]) {
		result = INT2NUM((int) qt_Smoke->ambiguousMethodList[i]);
//		PUSHs(sv_2mortal(newSViv(
//		    (IV)qt_Smoke->ambiguousMethodList[i]
//		)));
		i++;
	    }
	}
    }
    return result;
}

// findAllMethods(classid [, startingWith]) : returns { "mungedName" => [index in methods, ...], ... }

static VALUE
findAllMethods(int argc, VALUE * argv, VALUE self)
{
    VALUE classid = argv[0];
/*
    if(SvIOK(classid)) {
        Smoke::Index c = (Smoke::Index) SvIV(classid);
        char * pat = 0L;
        if(items > 1 && SvPOK(ST(1)))
            pat = SvPV_nolen(ST(1));
        Smoke::Index imax = qt_Smoke->numMethodMaps;
        Smoke::Index imin = 0, icur = -1, methmin, methmax;
        int icmp = -1;
        while(imax >= imin) {
            icur = (imin + imax) / 2;
            icmp = qt_Smoke->leg(qt_Smoke->methodMaps[icur].classId, c);
            if(!icmp) {
                Smoke::Index pos = icur;
                while(icur && qt_Smoke->methodMaps[icur-1].classId == c)
                    icur --;
                methmin = icur;
                icur = pos;
                while(icur < imax && qt_Smoke->methodMaps[icur+1].classId == c)
                    icur ++;
                methmax = icur;
                break;
            }
            if (icmp > 0)
		imax = icur - 1;
	    else
		imin = icur + 1;
        }
        if(!icmp) {
            for(Smoke::Index i=methmin ; i <= methmax ; i++) {
                Smoke::Index m = qt_Smoke->methodMaps[i].name;
                if(!pat || !strncmp(qt_Smoke->methodNames[m], pat, strlen(pat))) {
                    Smoke::Index ix= qt_Smoke->methodMaps[i].method;
                    AV* meths = newAV();
                    if(ix >= 0) {	// single match
                        av_push(meths, newSViv((IV)ix));
                    } else {		// multiple match
                        ix = -ix;		// turn into ambiguousMethodList index
                        while(qt_Smoke->ambiguousMethodList[ix]) {
                          av_push(meths, newSViv((IV)qt_Smoke->ambiguousMethodList[ix]));
                          ix++;
                        }
                    }
                    hv_store(return self;, qt_Smoke->methodNames[m],strlen(qt_Smoke->methodNames[m]),newRV_inc((SV*)meths),0);
                }
            }
        }
    }
    */
    return self;
}

static VALUE
dumpCandidates(VALUE self, VALUE rmeths)
{
/*
    if(SvROK(rmeths) && SvTYPE(SvRV(rmeths)) == SVt_PVAV) {
        AV *methods = (AV*)SvRV(rmeths);
        VALUE errmsg = rb_str_new2("");
        for(int i = 0; i <= av_len(methods); i++) {
                sv_catpv(errmsg, "\t");
                IV id = SvIV(*(av_fetch(methods, i, 0)));
                Smoke::Method &meth = qt_Smoke->methods[id];
                const char *tname = qt_Smoke->types[meth.ret].name;
                if(meth.flags & Smoke::mf_static) sv_catpv(errmsg, "static ");
                sv_catpvf(errmsg, "%s ", (tname ? tname:"void"));
                sv_catpvf(errmsg, "%s::%s(", qt_Smoke->classes[meth.classId].className, qt_Smoke->methodNames[meth.name]);
                for(int i = 0; i < meth.numArgs; i++) {
                        if(i) sv_catpv(errmsg, ", ");
                        tname = qt_Smoke->types[qt_Smoke->argumentList[meth.args+i]].name;
                        sv_catpv(errmsg, (tname ? tname:"void"));
                }
                sv_catpv(errmsg, ")");
                if(meth.flags & Smoke::mf_const) sv_catpv(errmsg, " const");
                sv_catpv(errmsg, "\n");
        }
        return self;=errmsg;
    }
    else
        return self;=rb_str_new2("");
   */
    return self;
}

static VALUE
rb_catArguments(VALUE self, VALUE r_args)
{
/*
    if(SvROK(r_args) && SvTYPE(SvRV(r_args)) == SVt_PVAV) {
        AV* args=(AV*)SvRV(r_args);
        for(int i = 0; i <= av_len(args); i++) {
            VALUE *arg=av_fetch(args, i, 0);
	    if(i) sv_catpv(return self;, ", ");
	    if(!arg || !SvOK(*arg)) {
		sv_catpv(return self;, "undef");
	    } else if(SvROK(*arg)) {
		smokeruby_object *o = value_obj_info(*arg);
		if(o)
		    sv_catpv(return self;, o->smoke->className(o->classId));
		else
		    sv_catsv(return self;, *arg);
	    } else {
		bool isString = SvPOK(*arg);
		STRLEN len;
		char *s = SvPV(*arg, len);
		if(isString) sv_catpv(return self;, "'");
		sv_catpvn(return self;, s, len > 10 ? 10 : len);
		if(len > 10) sv_catpv(return self;, "...");
		if(isString) sv_catpv(return self;, "'");
	    }
	}
    }
  */
    return self;
}

static VALUE
callMethod(int argc, VALUE * argv, VALUE self)
{
//    if(_current_method) {
//	MethodCall c(qt_Smoke, _current_method, &ST(0), items);
//	c.next();
//	VALUE ret = c.var();
//	SvREFCNT_inc(ret);
//	PUSHs(sv_2mortal(ret));
//    } else
//	PUSHs(sv_newmortal());
	return self;
}

static VALUE
isObject(VALUE self, VALUE obj)
{
	void * ptr = 0;
	ptr = value_to_ptr(obj);
    return (ptr > 0 ? Qtrue : Qfalse);
}

static VALUE
isRef(VALUE self, VALUE obj)
{
	VALUE klass = rb_funcall(obj, rb_intern("class"), 0);
	VALUE class_name = rb_funcall(klass, rb_intern("name"), 0);
	return strcmp(STR2CSTR(class_name), "Qt::Internal::Reference") == 0 ? Qtrue : Qfalse;
}

static VALUE
setCurrentMethod(VALUE self, VALUE meth_value)
{
    int meth = NUM2INT(meth_value);
    // FIXME: damn, this is lame, and it doesn't handle ambiguous methods
    _current_method = meth;  //qt_Smoke->methodMaps[meth].method;
    return self;
}

static VALUE
getClassList(VALUE self)
{
	VALUE class_list = rb_ary_new();

    for(int i = 1; i <= qt_Smoke->numClasses; i++) {
		rb_ary_push(class_list, rb_str_new2(qt_Smoke->classes[i].className));
    }
    
    return class_list;
}

static VALUE
installthis(VALUE self, VALUE package_value)
{
    char *package = STR2CSTR(package_value);
 /*
    if(!package) XSRETURN_EMPTY;
    char *name = new char[strlen(package) + 7];
    char *file = __FILE__;
    strcpy(name, package);
    strcat(name, "::this");
    // *{ $name } = sub () : lvalue;
    CV *thissub = newXS(name, XS_this, file);
    sv_setpv((SV*)thissub, "");    // sub this () : lvalue;
    delete[] name;
  */
	return self;
}

static VALUE
installattribute(VALUE self, VALUE package_value, VALUE name_value)
{
    char *package = STR2CSTR(package_value);
    char *name = STR2CSTR(name_value);
    if(!package || !name) return Qnil;
    char *attr = new char[strlen(package) + strlen(name) + 3];
    sprintf(attr, "%s::%s", package, name);
    char *file = __FILE__;
    // *{ $attr } = sub () : lvalue;
//    CV *attrsub = newXS(attr, XS_attr, file);
//    sv_setpv((SV*)attrsub, "");
//    CvLVALUE_on(attrsub);
//    CvNODEBUG_on(attrsub);
    delete[] attr;
    return self;
}

static VALUE
installsuper(VALUE self, VALUE package_value)
{
    char *package = STR2CSTR(package_value);
    if(!package) return Qnil;
    char *attr = new char[strlen(package) + 8];
    sprintf(attr, "%s::SUPER", package);
    char *file = __FILE__;
//    CV *attrsub = newXS(attr, XS_super, file);
//    sv_setpv((SV*)attrsub, "");
    delete[] attr;
    return self;
}

static VALUE
installautoload(VALUE self, VALUE package_value)
{
	char *package = STR2CSTR(package_value);
    if(	!package
    	|| strcmp(package, "Qt::t") == 0
    	|| strcmp(package, "Qt::Internal") == 0 )
    {
    	return Qnil;
    }

    VALUE klass = rb_define_class_under(qt_module, package+strlen("Qt::"), rb_cObject);
    rb_define_method(klass, "method_missing", (VALUE (*) (...)) method_missing, -1);
    rb_define_singleton_method(klass, "new", (VALUE (*) (...)) qt_new, -1);
    
//    char *autoload = new char[strlen(package) + 11];
//    strcpy(autoload, package);
//    strcat(autoload, "::_UTOLOAD");
 //   char *file = __FILE__;
    // *{ $package."::AUTOLOAD" } = XS_AUTOLOAD
//    newXS(autoload, XS_AUTOLOAD, file);
//    delete[] autoload;
    return klass;
}

static VALUE
ruby_this(VALUE self)
{
    return rv_this;
}

static VALUE
app(VALUE self)
{
    return rv_qapp;
}

static VALUE
version(VALUE self)
{
    return rb_str_new2(QT_VERSION_STR);
}

void
Init_Qt()
{
    init_qt_Smoke();
    qt_Smoke->binding = new QtRubySmokeBinding(qt_Smoke);
    install_handlers(Qt_handlers);

    methcache.setAutoDelete(1);
    classcache.setAutoDelete(1);

    qt_module = rb_define_module("Qt");
    qt_internal_module = rb_define_module_under(qt_module, "Internal");

	rb_define_method(qt_internal_module, "getMethStat", (VALUE (*) (...)) getMethStat, 0);
	rb_define_method(qt_internal_module, "getClassStat", (VALUE (*) (...)) getClassStat, 0);
	rb_define_method(qt_internal_module, "getIsa", (VALUE (*) (...)) getIsa, 1);
	rb_define_method(qt_internal_module, "dontRecurse", (VALUE (*) (...)) dontRecurse, 0);
	rb_define_method(qt_internal_module, "allocateMocArguments", (VALUE (*) (...)) allocateMocArguments, 1);
	rb_define_method(qt_internal_module, "setMocType", (VALUE (*) (...)) setMocType, 4);
	rb_define_method(qt_internal_module, "installsignal", (VALUE (*) (...)) installsignal, 1);
	rb_define_method(qt_internal_module, "installqt_invoke", (VALUE (*) (...)) installqt_invoke, 1);
#ifdef DEBUG
	rb_define_method(qt_internal_module, "setDebug", (VALUE (*) (...)) setDebug, 1);
	rb_define_method(qt_internal_module, "debug", (VALUE (*) (...)) debug, 0);
#endif
	rb_define_method(qt_internal_module, "getTypeNameOfArg", (VALUE (*) (...)) getTypeNameOfArg, 2);
	rb_define_method(qt_internal_module, "classIsa", (VALUE (*) (...)) classIsa, 2);
	rb_define_method(qt_internal_module, "insert_pclassid", (VALUE (*) (...)) insert_pclassid, 2);
	rb_define_method(qt_internal_module, "find_pclassid", (VALUE (*) (...)) find_pclassid, 1);
	rb_define_method(qt_internal_module, "insert_mcid", (VALUE (*) (...)) insert_mcid, 2);
	rb_define_method(qt_internal_module, "find_mcid", (VALUE (*) (...)) find_mcid, 1);
	rb_define_method(qt_internal_module, "getSVt", (VALUE (*) (...)) getSVt, 1);
	rb_define_method(qt_internal_module, "make_QUParameter", (VALUE (*) (...)) make_QUParameter, 4);
	rb_define_method(qt_internal_module, "make_QMetaData", (VALUE (*) (...)) make_QMetaData, 2);
	rb_define_method(qt_internal_module, "make_QUMethod", (VALUE (*) (...)) make_QUMethod, 2);
	rb_define_method(qt_internal_module, "make_QMetaData_tbl", (VALUE (*) (...)) make_QMetaData_tbl, 1);
	rb_define_method(qt_internal_module, "make_metaObject", (VALUE (*) (...)) make_metaObject, 6);
	rb_define_method(qt_internal_module, "dumpObjects", (VALUE (*) (...)) dumpObjects, 0);
	rb_define_method(qt_internal_module, "dangle", (VALUE (*) (...)) dangle, 1);
	rb_define_method(qt_internal_module, "setAllocated", (VALUE (*) (...)) setAllocated, 2);
	rb_define_method(qt_internal_module, "setqapp", (VALUE (*) (...)) setqapp, 1);
	rb_define_method(qt_internal_module, "setThis", (VALUE (*) (...)) setThis, 1);
	rb_define_method(qt_internal_module, "deleteObject", (VALUE (*) (...)) deleteObject, 1);
	rb_define_method(qt_internal_module, "mapObject", (VALUE (*) (...)) mapObject, 1);
	// isQOjbect => isaQObject
	rb_define_method(qt_internal_module, "isQObject", (VALUE (*) (...)) isaQObject, 1);
	rb_define_method(qt_internal_module, "isValidAllocatedPointer", (VALUE (*) (...)) isValidAllocatedPointer, 1);
	rb_define_method(qt_internal_module, "findAllocatedObjectFor", (VALUE (*) (...)) findAllocatedObjectFor, 1);
	rb_define_method(qt_internal_module, "idClass", (VALUE (*) (...)) idClass, 1);
	rb_define_method(qt_internal_module, "idMethodName", (VALUE (*) (...)) idMethodName, 1);
	rb_define_method(qt_internal_module, "idMethod", (VALUE (*) (...)) idMethod, 2);
	rb_define_method(qt_internal_module, "findMethod", (VALUE (*) (...)) findMethod, 2);
	rb_define_method(qt_internal_module, "findMethodFromIds", (VALUE (*) (...)) findMethodFromIds, 2);
	rb_define_method(qt_internal_module, "findAllMethods", (VALUE (*) (...)) findAllMethods, 2);
	rb_define_method(qt_internal_module, "dumpCandidates", (VALUE (*) (...)) dumpCandidates, 1);
	rb_define_method(qt_internal_module, "catArguments", (VALUE (*) (...)) catArguments, 1);
	rb_define_method(qt_internal_module, "callMethod", (VALUE (*) (...)) callMethod, -1);
	rb_define_method(qt_internal_module, "isObject", (VALUE (*) (...)) isObject, 1);
	rb_define_method(qt_internal_module, "isRef", (VALUE (*) (...)) isRef, 1);
	rb_define_method(qt_internal_module, "setCurrentMethod", (VALUE (*) (...)) setCurrentMethod, 1);
	rb_define_method(qt_internal_module, "getClassList", (VALUE (*) (...)) getClassList, 0);
	rb_define_method(qt_internal_module, "installthis", (VALUE (*) (...)) installthis, 1);
	rb_define_method(qt_internal_module, "installattribute", (VALUE (*) (...)) installattribute, 2);
	rb_define_method(qt_internal_module, "installsuper", (VALUE (*) (...)) installsuper, 1);
	rb_define_method(qt_internal_module, "installautoload", (VALUE (*) (...)) installautoload, 1);
	rb_define_method(qt_internal_module, "ruby_this", (VALUE (*) (...)) ruby_this, 0);
	rb_define_method(qt_internal_module, "app", (VALUE (*) (...)) app, 0);
	rb_define_method(qt_internal_module, "version", (VALUE (*) (...)) version, 0);

	pointer_map = rb_hash_new();
	rb_gc_register_address(&pointer_map);
//	rv_this = ??;
//	rb_gc_register_address(&rv_this);

	rb_include_module(qt_module, qt_internal_module);
	rb_require("qtruby.rb");

	// TODO: Is there a ruby function to replace this rb_eval_string() call?
	rb_eval_string("include Qt");

	// Do package initialization
	rb_funcall(qt_internal_module, rb_intern("init"), 0);
}

};
